/*
 * x86 performance-monitoring counters driver.
 *
 * Copyright (C) 1999-2000  Mikael Pettersson
 */
#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/perfctr.h>

#include <asm/msr.h>

#include "compat.h"
#include "x86_compat.h"

/****************************************************************
 *								*
 * Hardware-level declarations.					*
 *								*
 ****************************************************************/

static union {	/* Support for lazy evntsel MSR updates. */
	struct perfctr_control control;	/* 16 bytes */
	char __align[SMP_CACHE_BYTES];	/* 32 bytes */
} per_cpu_control[NR_CPUS] __cacheline_aligned;

/* Intel P5, Cyrix 6x86MX/MII/III, Centaur WinChip C6/2/2A/3 */
#define MSR_P5_CESR		0x11
#define MSR_P5_CTR0		0x12
#define MSR_P5_CTR1		0x13

/* Intel P6 */
#define MSR_P6_PERFCTR0		0xC1
#define MSR_P6_PERFCTR1		0xC2
#define MSR_P6_EVNTSEL0		0x186
#define MSR_P6_EVNTSEL1		0x187

/* AMD K7 Athlon */
#define MSR_K7_EVNTSEL0		0xC0010000	/* .. 0xC0010003 */
#define MSR_K7_PERFCTR0		0xC0010004	/* .. 0xC0010007 */

static __inline__ unsigned int get_cr4(void)
{
	unsigned int cr4;
	__asm__ __volatile__("movl %%cr4,%0" : "=r"(cr4));
	return cr4;
}

#define rdmsrl(msr,low) \
	__asm__ __volatile__("rdmsr" : "=a"(low) : "c"(msr) : "edx")
#define rdpmcl(ctr,low) \
	__asm__ __volatile__("rdpmc" : "=a"(low) : "c"(ctr) : "edx")

/****************************************************************
 *								*
 * Processor-specific procedures and values.			*
 *								*
 ****************************************************************/

/* Detected CPU type. */
unsigned char perfctr_cpu_type;

/* CPU features optionally supported.
 * RDPMC and RDTSC are on by default. They will be disabled
 * by the init procedures if necessary.
 */
unsigned char perfctr_cpu_features
	= PERFCTR_FEATURE_RDPMC | PERFCTR_FEATURE_RDTSC;

static void redirect_call(void *ra, void *to)
{
	if( *((unsigned char*)ra - 5) != 0xE8 ) {
		printk(KERN_INFO __FILE__ ":" __FUNCTION__
		       ": unable to redirect caller %p to %p\n",
		       ra, to);
		return;
	}
	*(unsigned int*)((char*)ra - 4) = (char*)to - (char*)ra;
}

static void (*read_counters)(int, struct perfctr_low_ctrs*);
void perfctr_cpu_read_counters(int nrctrs, struct perfctr_low_ctrs *ctrs)
{
	redirect_call(__builtin_return_address(0), read_counters);
	return read_counters(nrctrs, ctrs);
}

static void (*write_control)(int nrctrs, const struct perfctr_control *control);
void perfctr_cpu_write_control(int nrctrs, const struct perfctr_control *control)
{
	redirect_call(__builtin_return_address(0), write_control);
	return write_control(nrctrs, control);
}

static int (*check_control)(const struct perfctr_control *control);
int perfctr_cpu_check_control(const struct perfctr_control *control)
{
	redirect_call(__builtin_return_address(0), check_control);
	return check_control(control);
}

static void p5_read_counters(int nrctrs, struct perfctr_low_ctrs *ctrs)
{
	rdtscl(ctrs->tsc);
	rdmsrl(MSR_P5_CTR0, ctrs->pmc[0]);
	if( nrctrs == 2 )
		rdmsrl(MSR_P5_CTR1, ctrs->pmc[1]);
}

static void p5mmx_read_counters(int nrctrs, struct perfctr_low_ctrs *ctrs)
{
	rdtscl(ctrs->tsc);
	rdpmcl(0, ctrs->pmc[0]);
	if( nrctrs == 2 )
		rdpmcl(1, ctrs->pmc[1]);
}

static void p5_write_control(int nrctrs, const struct perfctr_control *control)
{
	struct perfctr_control *cpu;
	unsigned evntsel;

	cpu = &per_cpu_control[smp_processor_id()].control;
	evntsel = control->evntsel[0];
	if( cpu->evntsel[0] != evntsel ) {
		cpu->evntsel[0] = evntsel;
		wrmsr(MSR_P5_CESR, evntsel, 0);
	}
}

static int p5_check_control(const struct perfctr_control *control)
{
	/* protect reserved and pin control bits */
	if( control->evntsel[0] & 0xFE00FE00 )
		return -EPERM;
	/* CTR1 is on if its CPL field is non-zero */
	if( control->evntsel[0] & 0x00C00000 )
		return 2;
	/* CTR0 is on if its CPL field is non-zero */
	if( control->evntsel[0] & 0x000000C0 )
		return 1;
	return 0;
}

static void p6_write_control(int nrctrs, const struct perfctr_control *control)
{
	struct perfctr_control *cpu;
	unsigned evntsel;

	cpu = &per_cpu_control[smp_processor_id()].control;
	if( nrctrs == 2 ) {
		evntsel = control->evntsel[1];
		if( evntsel != cpu->evntsel[1] ) {
			cpu->evntsel[1] = evntsel;
			wrmsr(MSR_P6_EVNTSEL1, evntsel, 0);
		}
	}	
	evntsel = control->evntsel[0];
	if( evntsel != cpu->evntsel[0] ) {
		cpu->evntsel[0] = evntsel;
		wrmsr(MSR_P6_EVNTSEL0, evntsel, 0);
	}
}

static int p6_check_control(const struct perfctr_control *control)
{
	/* protect reserved, interrupt control, and pin control bits */
	if( control->evntsel[0] & 0x00380000 )
		return -EPERM;
	if( control->evntsel[1] & 0x00780000 )
		return -EPERM;
	/* check global enable bit */
	if( control->evntsel[0] & 0x00400000 ) {
		/* check CPL field */
		if( control->evntsel[1] & 0x00030000 )
			return 2;
		if( control->evntsel[0] & 0x00030000 )
			return 1;
	}
	return 0;
}

static void k7_read_counters(int nrctrs, struct perfctr_low_ctrs *ctrs)
{
	int i;

	rdtscl(ctrs->tsc);
	for(i = nrctrs; --i >= 0;)
		rdpmcl(i, ctrs->pmc[i]);
}

static void k7_write_control(int nrctrs, const struct perfctr_control *control)
{
	struct perfctr_control *cpu;
	int i;
	unsigned evntsel;

	cpu = &per_cpu_control[smp_processor_id()].control;
	for(i = nrctrs; --i >= 0;) {
		evntsel = control->evntsel[i];
		if( evntsel != cpu->evntsel[i] ) {
			cpu->evntsel[i] = evntsel;
			wrmsr(MSR_K7_EVNTSEL0+i, evntsel, 0);
		}
	}
}

static int k7_check_control(const struct perfctr_control *control)
{
	int i, last_on;

	last_on = -1;
	for(i = 0; i < 4; ++i) {
		/* protect reserved, interrupt control, and pin control bits */
		if( control->evntsel[i] & 0x00380000 )
			return -EPERM;
		/* check enable bit */
		if( !(control->evntsel[i] & 0x00400000) )
			continue;
		/* check CPL field */
		if( control->evntsel[i] & 0x00030000 )
			last_on = i;
	}
	return last_on + 1;
}

static int mii_check_control(const struct perfctr_control *control)
{
	/* Protect reserved and pin control bits.
	 * CESR bits 9 and 25 are reserved in the Cyrix III,
	 * but not in the earlier processors where they control
	 * the external PM pins. In either case, we do not
	 * allow the user to touch them.
	 */
	if( control->evntsel[0] & 0xFA00FA00 )
		return -EPERM;
	/* CTR1 is on if its CPL field is non-zero */
	if( control->evntsel[0] & 0x00C00000 )
		return 2;
	/* CTR0 is on if its CPL field is non-zero */
	if( control->evntsel[0] & 0x000000C0 )
		return 1;
	return 0;
}

#if defined(CONFIG_PERFCTR_WINCHIP)
static void c6_read_counters(int nrctrs, struct perfctr_low_ctrs *ctrs)
{
	++ctrs->tsc;
	rdpmcl(0, ctrs->pmc[0]);
	rdpmcl(1, ctrs->pmc[1]);
}

static int c6_check_control(const struct perfctr_control *control)
{
	/* protect reserved bits */
	if( control->evntsel[0] & 0xFF00FF00 )
		return -EPERM;
	/* we have to assume both counters are active */
	return 2;
}
#endif	/* CONFIG_PERFCTR_WINCHIP */

/****************************************************************
 *								*
 * Initialisation procedures.					*
 *								*
 ****************************************************************/

#ifdef CONFIG_PERFCTR_INIT_TESTS

#define NITER	64
#define X2(S)	S";"S
#define X8(S)	X2(X2(X2(S)))

static void __init do_rdpmc(unsigned unused1, unsigned unused2)
{
	unsigned i;
	for(i = 0; i < NITER/8; ++i)
		__asm__ __volatile__(X8("rdpmc") : : "c"(0) : "eax", "edx");
}

static void __init do_rdmsr(unsigned msr, unsigned unused2)
{
	unsigned i;
	for(i = 0; i < NITER/8; ++i)
		__asm__ __volatile__(X8("rdmsr") : : "c"(msr) : "eax", "edx");
}

static void __init do_wrmsr(unsigned msr, unsigned data)
{
	unsigned i;
	for(i = 0; i < NITER/8; ++i)
		__asm__ __volatile__(X8("wrmsr") : : "c"(msr), "a"(data), "d"(0));
}

static void __init do_rdcr4(unsigned unused1, unsigned unused2)
{
	unsigned i;
	unsigned dummy;
	for(i = 0; i < NITER/8; ++i)
		__asm__ __volatile__(X8("movl %%cr4,%0") : "=r"(dummy));
}

static void __init do_wrcr4(unsigned cr4, unsigned unused2)
{
	unsigned i;
	for(i = 0; i < NITER/8; ++i)
		__asm__ __volatile__(X8("movl %0,%%cr4") : : "r"(cr4));
}

static void __init do_empty_loop(unsigned unused1, unsigned unused2)
{
	unsigned i;
	for(i = 0; i < NITER/8; ++i)
		__asm__ __volatile__("" : : "c"(0));
}

static unsigned __init run(void (*doit)(unsigned, unsigned),
			   unsigned arg1, unsigned arg2)
{
	unsigned start, dummy, stop;
	rdtsc(start, dummy);
	(*doit)(arg1, arg2);	/* should take < 2^32 cycles to complete */
	rdtsc(stop, dummy);
	return stop - start;
}

static struct {
	const char *name;
	unsigned ticks;
} access_methods[] __initdata = {
	{ "rdpmc" },
	{ "rdmsr (counter)" },
	{ "rdmsr (evntsel)" },
	{ "wrmsr (counter)" },
	{ "wrmsr (evntsel)" },
	{ "read %cr4" },
	{ "write %cr4" },
	{ "loop overhead" },
};

static void __init init_tests_message(void)
{
	printk("Please email the following PERFCTR INIT lines "
	       "to mikpe@csd.uu.se\n"
	       "To remove this message, rebuild the kernel "
	       "with CONFIG_PERFCTR_INIT_TESTS=n\n");
	printk("PERFCTR INIT: vendor %u, family %u, model %u\n",
	       boot_cpu_data.x86_vendor,
	       boot_cpu_data.x86,
	       boot_cpu_data.x86_model);
}

static void __init
measure_overheads(unsigned msr_evntsel0, unsigned evntsel0, unsigned msr_perfctr0)
{
	int i;

	access_methods[0].ticks =
		(perfctr_cpu_features & PERFCTR_FEATURE_RDPMC)
		? run(do_rdpmc,0,0)
		: 0;
	access_methods[1].ticks = run(do_rdmsr, msr_perfctr0, 0);
	access_methods[2].ticks = run(do_rdmsr, msr_evntsel0, 0);
	access_methods[3].ticks = run(do_wrmsr, msr_perfctr0, 0);
	access_methods[4].ticks = run(do_wrmsr, msr_evntsel0, evntsel0);
	access_methods[5].ticks = run(do_rdcr4, 0, 0);
	access_methods[6].ticks = run(do_wrcr4, get_cr4(), 0);
	access_methods[7].ticks = run(do_empty_loop, 0, 0);
	wrmsr(msr_evntsel0, 0, 0);

	init_tests_message();
	printk("PERFCTR INIT: NITER == %u\n", NITER);
	for(i = 0; i < ARRAY_SIZE(access_methods); ++i) {
		printk("PERFCTR INIT: %s ticks == %u\n",
		       access_methods[i].name, access_methods[i].ticks);
	}
}

static void __init p5_init_tests(void)
{
	unsigned evnt = 0x16 | (3 << 6);
	measure_overheads(MSR_P5_CESR, evnt, MSR_P5_CTR0);
}

static void __init p6_init_tests(void)
{
	unsigned evnt = 0xC0 | (3 << 16) | (1 << 22);
	measure_overheads(MSR_P6_EVNTSEL0, evnt, MSR_P6_PERFCTR0);
}

static void __init k7_stop_and_clear(void)
{
	wrmsr(MSR_K7_EVNTSEL0+0, 0, 0);
	wrmsr(MSR_K7_EVNTSEL0+1, 0, 0);
	wrmsr(MSR_K7_PERFCTR0+0, 0, 0);
	wrmsr(MSR_K7_PERFCTR0+1, 0, 0);
}

static unsigned __init k7_test(unsigned ctr0_on, unsigned ctr1_on)
{
	unsigned ctr1_evnt = 0xC0 | (3 << 16);	/* num insns */
	unsigned ctr0_evnt = 0xDF | (3 << 16);	/* dr3 match */
	const unsigned ENable = (1<<22);
	unsigned dummy, ctr1_result;

	if( ctr1_on ) ctr1_evnt |= ENable;
	if( ctr0_on ) ctr0_evnt |= ENable;
	wrmsr(MSR_K7_EVNTSEL0+1, ctr1_evnt, 0);
	wrmsr(MSR_K7_EVNTSEL0+0, ctr0_evnt, 0);
	do_empty_loop(0, 0);
	rdmsr(MSR_K7_PERFCTR0+1, ctr1_result, dummy);
	k7_stop_and_clear();
	return ctr1_result;
}

static void __init k7_check_how_ENable_works(void)
{
	unsigned test0, test1, test2, test3;

	k7_stop_and_clear();

	test0 = k7_test(0, 0); /* should be == 0 */
	test1 = k7_test(1, 1); /* should be > 0 */
	test2 = k7_test(0, 1); /* if > 0, EvntSel0 does not override */
	test3 = k7_test(1, 0); /* if > 0, only EvntSel0 has ENable */

	printk("PERFCTR INIT: Athlon test0 == %u (%s)\n",
	       test0, test0 == 0 ? "ok" : "unexpected");
	printk("PERFCTR INIT: Athlon test1 == %u (%s)\n",
	       test1, test1 > 0 ? "ok" : "unexpected");
	printk("PERFCTR INIT: Athlon test2 == %u (%s)\n",
	       test2, test2 > 0 ? "EvntSel0 does not override" : "ok");
	printk("PERFCTR INIT: Athlon test3 == %u (%s)\n",
	       test3, test3 > 0 ? "only EvntSel0 has ENable" : "ok");
}

static void __init k7_init_tests(void)
{
	unsigned evnt = 0xC0 | (3 << 16) | (1 << 22);
	measure_overheads(MSR_K7_EVNTSEL0, evnt, MSR_K7_PERFCTR0);
	k7_check_how_ENable_works();
}

#define mii_init_tests()	p5_init_tests()

#if defined(CONFIG_PERFCTR_WINCHIP)
static void __init c6_init_tests(void)
{
	unsigned int cesr, dummy;

	rdmsr(MSR_P5_CESR, cesr, dummy);
	init_tests_message();
	printk("PERFCTR INIT: boot CESR == %#08x\n", cesr);
}
#endif

#else	/* !CONFIG_PERFCTR_INIT_TESTS */
#define p5_init_tests()
#define p6_init_tests()
#define k7_init_tests()
#define mii_init_tests()
#define c6_init_tests()
#endif	/* CONFIG_PERFCTR_INIT_TESTS */

static int __init intel_init(void)
{
	if( !(boot_cpu_data.x86_capability & X86_FEATURE_TSC) )
		return -ENODEV;
	switch( boot_cpu_data.x86 ) {
	case 5:
		if( boot_cpu_data.x86_capability & X86_FEATURE_MMX ) {
			perfctr_cpu_type = PERFCTR_CPU_INTEL_P5MMX;
			read_counters = p5mmx_read_counters;
		} else {
			perfctr_cpu_type = PERFCTR_CPU_INTEL_P5;
			perfctr_cpu_features &= ~PERFCTR_FEATURE_RDPMC;
			read_counters = p5_read_counters;
		}
		write_control = p5_write_control;
		check_control = p5_check_control;
		p5_init_tests();
		return 0;
	case 6:
		if( boot_cpu_data.x86_model >= 7 )	/* PIII */
			perfctr_cpu_type = PERFCTR_CPU_INTEL_PIII;
		else if( boot_cpu_data.x86_model >= 3 )	/* PII or Celeron */
			perfctr_cpu_type = PERFCTR_CPU_INTEL_PII;
		else
			perfctr_cpu_type = PERFCTR_CPU_INTEL_P6;
		read_counters = p5mmx_read_counters;
		write_control = p6_write_control;
		check_control = p6_check_control;
		p6_init_tests();
		return 0;
	}
	return -ENODEV;
}

static int __init amd_init(void)
{
	if( !(boot_cpu_data.x86_capability & X86_FEATURE_TSC) )
		return -ENODEV;
	switch( boot_cpu_data.x86 ) {
	case 6:	/* K7 Athlon
		 * Model 2 adds local APIC.
		 */
		perfctr_cpu_type = PERFCTR_CPU_AMD_K7;
		read_counters = k7_read_counters;
		write_control = k7_write_control;
		check_control = k7_check_control;
		k7_init_tests();
		return 0;
	}
	return -ENODEV;
}

static int __init cyrix_init(void)
{
	if( !(boot_cpu_data.x86_capability & X86_FEATURE_TSC) )
		return -ENODEV;
	switch( boot_cpu_data.x86 ) {
	case 6:	/* 6x86MX, MII, or III */
		perfctr_cpu_type = PERFCTR_CPU_CYRIX_MII;
		read_counters = p5mmx_read_counters;
		write_control = p5_write_control;
		check_control = mii_check_control;
		mii_init_tests();
		return 0;
	}
	return -ENODEV;
}

static int __init centaur_init(void)
{
#if defined(CONFIG_PERFCTR_WINCHIP)
	switch( boot_cpu_data.x86 ) {
	case 5:
		switch( boot_cpu_data.x86_model ) {
		case 4: /* WinChip C6 */
			perfctr_cpu_type = PERFCTR_CPU_WINCHIP_C6;
			break;
		case 8: /* WinChip 2 or 2A */
		case 9: /* WinChip 3, a 2A with 128KB L1 cache */
			perfctr_cpu_type = PERFCTR_CPU_WINCHIP_2;
			break;
		default:
			return -ENODEV;
		}
		/*
		 * TSC must be inaccessible for perfctrs to work.
		 */
		if( !(get_cr4() & X86_CR4_TSD) ||
		    (boot_cpu_data.x86_capability & X86_FEATURE_TSC) )
			return -ENODEV;
		perfctr_cpu_features &= ~PERFCTR_FEATURE_RDTSC;
		read_counters = c6_read_counters;
		write_control = p5_write_control;
		check_control = c6_check_control;
		c6_init_tests();
		return 0;
	}
#endif
	return -ENODEV;
}

int __init perfctr_cpu_init(void)
{
	int err = -ENODEV;
	if( boot_cpu_data.x86_capability & X86_FEATURE_MSR ) {
		switch( boot_cpu_data.x86_vendor ) {
		case X86_VENDOR_INTEL:
			err = intel_init();
			break;
		case X86_VENDOR_AMD:
			err = amd_init();
			break;
		case X86_VENDOR_CYRIX:
			err = cyrix_init();
			break;
		case X86_VENDOR_CENTAUR:
			err = centaur_init();
		}
	}
	if( err )
		return err;
	/*
	 * per_cpu_control[] is initialised to contain "impossible"
	 * evntsel values guaranteed to differ from anything accepted
	 * by perfctr_cpu_check_control(). This way, initialisation of
	 * a CPU's evntsel MSRs will happen automatically the first time
	 * perfctr_cpu_write_control() executes on it.
	 * All-bits-one works for all currently supported processors.
	 * [XXX: Should be set up by the cpu-specific init procedure.]
	 */
	memset(per_cpu_control, ~0, sizeof per_cpu_control);

	/*
	 * Set CR4.PCE, if the processor supports user-mode RDPMC.
	 * CR4.PCE is enabled globally, for two reasons:
	 *
	 * 1. Writing %cr4 to toggle CR4.PCE at process suspend/resume
	 *    is expensive (a Pentium II/III needs 42 cycles just to
	 *    write to %cr4).
	 *
	 * 2. Starting with the 2.3.40 kernel, %cr4 is written to from
	 *    a single global variable mmu_cr4_features in post-boot
	 *    operation. A change to CR4.PCE on one CPU would be undone
	 *    asynchronously whenever flush_tlb_all() is called.
	 */
	if( perfctr_cpu_features & PERFCTR_FEATURE_RDPMC )
		set_cr4_pce_global();

	return 0;
}

void __exit perfctr_cpu_exit(void)
{
	clear_cr4_pce_global();
}

/****************************************************************
 *								*
 * Hardware reservation.					*
 *								*
 ****************************************************************/

static const char *current_service = 0;

const char *perfctr_cpu_reserve(const char *service)
{
	if( current_service )
		return current_service;
	current_service = service;
	return 0;
}

void perfctr_cpu_release(const char *service)
{
	if( service != current_service ) {
		printk(KERN_ERR __FUNCTION__
		       ": attempt by %s to release while reserved by %s\n",
		       service, current_service);
	} else
		current_service = 0;
}
